/*         ______   ___    ___ 
 *        /\  _  \ /\_ \  /\_ \ 
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *      By Shawn Hargreaves,
 *      1 Salisbury Road,
 *      Market Drayton,
 *      Shropshire,
 *      England, TF9 1AJ.
 *
 *      The 3D scene renderer.
 *
 *      See readme.txt for copyright information.
 */


#include <float.h>
#include <sys/farptr.h>

#include <allegro.h>
#include "p3dpriv.h"
#include "p3d.h"


static POLYGON_EDGE *scene_edge = NULL, *scene_inact;
static POLYGON_INFO *scene_poly = NULL;
static int scene_nedge = 0, scene_maxedge = 0;
static int scene_npoly = 0, scene_maxpoly = 0;
static BITMAP *scene_bmp;
static COLOR_MAP *scene_cmap;
static int scene_alpha;
static int last_x, scene_y, scene_addr;
static float last_z;
float scene_gap = 100.0;


SCANLINE_FILLER dummy_filler;
void dummy_func()
{
   dummy_filler = _poly_scanline_gcol8;
}


/* add_edge:
 *  Adds an edge structure to a linked list, returning the new head pointer.
 */
static POLYGON_EDGE *add_edge(POLYGON_EDGE *list, POLYGON_EDGE *edge, int sort_by_x)
{
   POLYGON_EDGE *pos = list;
   POLYGON_EDGE *prev = NULL;

   if (sort_by_x) {
      while ((pos) && ((pos->x + pos->dx / 2) <
		       (edge->x + edge->dx / 2))) {
	 prev = pos;
	 pos = pos->next;
      }
   }
   else {
      while ((pos) && (pos->top < edge->top)) {
	 prev = pos;
	 pos = pos->next;
      }
   }

   edge->next = pos;
   edge->prev = prev;

   if (pos)
      pos->prev = edge;

   if (prev) {
      prev->next = edge;
      return list;
   }
   else
      return edge;
}



/* remove_edge:
 *  Removes an edge structure from a list, returning the new head pointer.
 */
static POLYGON_EDGE *remove_edge(POLYGON_EDGE *list, POLYGON_EDGE *edge)
{
   if (edge->next) 
      edge->next->prev = edge->prev;

   if (edge->prev) {
      edge->prev->next = edge->next;
      return list;
   }
   else
      return edge->next;
}



/* scene_start:
 * Initializes a scene. The bitmap is the bitmap you will eventually render
 * on.
 * nedge and npoly are your estimate of how many edges and how many
 * polygons you will render (you cannot get over the limit specified here).
 * If you use same values in succesive calls, the space will be reused.
 * If you call with nedge=0 or npoly=0, the space is freed.
 * Returns negative numbers if allocations fail, zero on success.
 */
int scene_start(BITMAP *bmp, int nedge, int npoly)
{
   if (nedge == 0 || npoly == 0) {
      scene_maxedge = scene_maxpoly = 0;
      if (scene_nedge) {
         free(scene_edge);
         scene_edge = NULL;
      }
      if (scene_npoly) {
         free(scene_poly);
         scene_poly = NULL;
      }
      return 0;
   }

   if (nedge != scene_maxedge) {        /* don't realloc, we don't need */
      scene_maxedge = 0;                /* the old data */
      if (scene_edge) free(scene_edge);
      scene_edge = malloc(nedge * sizeof(POLYGON_EDGE));
      if (!scene_edge) return -1;
   }
   if (npoly != scene_maxpoly) {
      scene_maxpoly = 0;
      if (scene_poly) free(scene_poly);
      scene_poly = malloc(npoly * sizeof(POLYGON_INFO));
      if (!scene_poly) {
         free(scene_edge);
         scene_edge = NULL;
         return -2;
      }
   }
   scene_inact = NULL;
   scene_bmp = bmp;
   scene_maxedge = nedge;
   scene_maxpoly = npoly;
   scene_nedge = scene_npoly = 0;
   return 0;
}


/* poly_plane:
 * Computes the plane's equation coefficients A, B, C, D, in the original
 * space (before projection) for perspective projection:
 *   Axz + Byz + Cz + D = 0
 */
static void poly_plane(V3D *vtx[], POLYGON_INFO *poly)
{
   float x1, y1, z1, x2, y2, z2, x3, y3, z3, d;

   z1 = fixtof(vtx[0]->z);
   x1 = fixtof(vtx[0]->x) * z1; y1 = fixtof(vtx[0]->y) * z1;
   z2 = fixtof(vtx[1]->z);
   x2 = fixtof(vtx[1]->x) * z2; y2 = fixtof(vtx[1]->y) * z2;
   z3 = fixtof(vtx[2]->z);
   x3 = fixtof(vtx[2]->x) * z3; y3 = fixtof(vtx[2]->y) * z3;

   poly->a = y1*(z2 - z3) + y2*(z3 - z1) + y3*(z1 - z2);
   poly->b = x1*(z3 - z2) + x2*(z1 - z3) + x3*(z2 - z1);
   poly->c = x1*(y2 - y3) + x2*(y3 - y1) + x3*(y1 - y2);
   d = -(x1*(y3*z2 - y2*z3) + x2*(y1*z3 - y3*z1) + x3*(y2*z1 - y1*z2));
   /* prepare A, B, C for extracting 1/z */
   if ((d < 1e-10) && (d > -1e-10)) {
      if (d >= 0) d = 1e-10;
      else d = -1e-10;
   }
   poly->a /= d; poly->b /= d; poly->c /= d;
}

/* poly_plane_f:
 * Floating point version of poly_plane.
 */
static void poly_plane_f(V3D_f *vtx[], POLYGON_INFO *poly)
{
   float x1, y1, z1, x2, y2, z2, x3, y3, z3, d;

   z1 = vtx[0]->z;
   x1 = vtx[0]->x * z1; y1 = vtx[0]->y * z1;
   z2 = vtx[1]->z;
   x2 = vtx[1]->x * z2; y2 = vtx[1]->y * z2;
   z3 = vtx[2]->z;
   x3 = vtx[2]->x * z3; y3 = vtx[2]->y * z3;

   poly->a = y1*(z2 - z3) + y2*(z3 - z1) + y3*(z1 - z2);
   poly->b = x1*(z3 - z2) + x2*(z1 - z3) + x3*(z2 - z1);
   poly->c = x1*(y2 - y3) + x2*(y3 - y1) + x3*(y1 - y2);
   d = -(x1*(y3*z2 - y2*z3) + x2*(y1*z3 - y3*z1) + x3*(y2*z1 - y1*z2));
   /* prepare A, B, C for extracting 1/z */
   if ((d < 1e-10) && (d > -1e-10)) {
      if (d >= 0) d = 1e-10;
      else d = -1e-10;
   }
   poly->a /= d; poly->b /= d; poly->c /= d;
}



/* plane_z:
 * Returns 1/z of the plane in the point x, y.
 */
static float plane_z(int x, int y, POLYGON_INFO *poly)
{
   return poly->a * x + poly->b * y + poly->c;
}


/* far_z:
 * Compares z values of the plane to which 'edge' belongs and the plane
 * of 'poly', in point x, y.
 * Returns true if 'edge' is farther than 'poly'.
 */
static int far_z(int x, int y, POLYGON_EDGE *edge, POLYGON_INFO *poly)
{
   float z1, z2, zd;

   z1 = edge->dat.z;
   z2 = plane_z(x, y, poly);
   zd = (z2 - z1) * scene_gap;
   if (zd > z1) return 1;
   if (zd < -z1) return 0;

   return (z2 - z1 + 100.0 * (poly->a - edge->poly->a) > 0);
}



/* plane_z:
 * Initializes different elements in the poly structure.
 */
static void init_poly(int type, POLYGON_INFO *poly)
{
   static int flag_table[] = {
      INTERP_Z,
      INTERP_Z,
      INTERP_Z,
      INTERP_Z,
      INTERP_Z,
      INTERP_Z | INTERP_THRU,
      INTERP_Z | INTERP_THRU,
      INTERP_Z | INTERP_BLEND,
      INTERP_Z | INTERP_BLEND,
      INTERP_Z | INTERP_THRU | INTERP_BLEND,
      INTERP_Z | INTERP_THRU | INTERP_BLEND,
      INTERP_Z | INTERP_THRU | INTERP_TRANS,
      INTERP_Z | INTERP_THRU | INTERP_TRANS,
      INTERP_Z | INTERP_THRU | INTERP_TRANS,
      INTERP_Z | INTERP_THRU | INTERP_TRANS
   };
   
   poly->alt_drawer = _optim_alternative_drawer;
   poly->inside = 0;

   type &= ~POLYTYPE_ZBUF;
   poly->flags |= flag_table[type];

   if (poly->flags & POLYTYPE_ZBUF) {
      poly->flags |= INTERP_THRU;
   }
   if ((type == POLYTYPE_FLAT) && (_drawing_mode == DRAW_MODE_TRANS)) {
      poly->flags |= INTERP_THRU | INTERP_TRANS;
   }

   poly->cmap = color_map;
   poly->alpha = _blender_alpha;
   if (poly->flags & INTERP_BLEND) {
      if (bitmap_color_depth(scene_bmp) == 8) {
         poly->flags &= ~INTERP_BLEND;
      } else {
         poly->b15 = _blender_col_15;
         poly->b16 = _blender_col_16;
         poly->b24 = _blender_col_24;
         poly->b32 = _blender_col_32;
      }
   }

   scene_npoly++;
}



/* scene_polygon3d:
 * Put a polygon in the rendering list. Nothing is really rendered at this
 * moment. Should be called between scene_start() and scene_render().
 * Unlike polygon3d(), the polygon may be concave or self-intersecting.
 * Shapes that penetrate one another may look OK, but they are not really
 * handled by this code.
 * Note that the texture is stored as a pointer only, and you should keep
 * the actual bitmap until scene_render(), where it is used.
 * Returns zero on success, negative numbers if there is no space in the
 * edge or polygon tables, positive if it won't be rendered for lack of
 * rendering routine.
 */
int scene_polygon3d(int type, BITMAP *texture, int vc, V3D *vtx[])
{
   int c;
   int v1y, v2y;
   V3D *v1, *v2;
   POLYGON_EDGE *edge;
   POLYGON_INFO *poly;

   if (scene_nedge + vc > scene_maxedge) return -1;
   if (scene_npoly >= scene_maxpoly) return -2;

   edge = &scene_edge[scene_nedge];
   poly = &scene_poly[scene_npoly];

   /* set up the drawing mode */
   poly->drawer = _get_scanline_filler(type, &poly->flags, &poly->info,
                                      texture, scene_bmp);
   if (!poly->drawer) return 1;

   init_poly(type, poly);
   poly->color = vtx[0]->c;
   poly_plane(vtx, poly);

   v2 = vtx[vc-1];

   /* fill the edge table */
   for (c=0; c<vc; c++) {
      v1 = v2;
      v2 = vtx[c];
      v1y = fixtoi(v1->y);
      v2y = fixtoi(v2->y);

      if ((v1y != v2y) &&
	  ((v1y >= scene_bmp->ct) || (v2y >= scene_bmp->ct)) &&
	  ((v1y < scene_bmp->cb) || (v2y < scene_bmp->cb))) {

	 _fill_3d_edge_structure(edge, v1, v2, poly->flags, scene_bmp);
         edge->poly = poly;

	 if (edge->top < scene_bmp->ct) {
	    int gap = scene_bmp->ct - edge->top;
	    edge->top = scene_bmp->ct;
	    edge->x += edge->dx * gap;
	    _clip_polygon_segment(&edge->dat, gap, poly->flags);
	 }

	 if (edge->bottom >= edge->top) {
	    scene_inact = add_edge(scene_inact, edge, FALSE);
	    edge++;
            scene_nedge++;
	 }
      }
   }
   return 0;
}

/* scene_polygon3d_f:
 * Floating point version of scene_polygon3d.
 */
int scene_polygon3d_f(int type, BITMAP *texture, int vc, V3D_f *vtx[])
{
   int c;
   int v1y, v2y;
   V3D_f *v1, *v2;
   POLYGON_EDGE *edge;
   POLYGON_INFO *poly;

   if (scene_nedge + vc > scene_maxedge) return -1;
   if (scene_npoly >= scene_maxpoly) return -2;

   edge = &scene_edge[scene_nedge];
   poly = &scene_poly[scene_npoly];

   /* set up the drawing mode */
   poly->drawer = _get_scanline_filler(type, &poly->flags, &poly->info,
                                      texture, scene_bmp);
   if (!poly->drawer) return 1;

   init_poly(type, poly);
   poly->color = vtx[0]->c;
   poly_plane_f(vtx, poly);

   v2 = vtx[vc-1];

   /* fill the edge table */
   for (c=0; c<vc; c++) {
      v1 = v2;
      v2 = vtx[c];
      v1y = (int)(v1->y+0.5);
      v2y = (int)(v2->y+0.5);

      if ((v1y != v2y) &&
	  ((v1y >= scene_bmp->ct) || (v2y >= scene_bmp->ct)) &&
	  ((v1y < scene_bmp->cb) || (v2y < scene_bmp->cb))) {

	 _fill_3d_edge_structure_f(edge, v1, v2, poly->flags, scene_bmp);
         edge->poly = poly;

	 if (edge->top < scene_bmp->ct) {
	    int gap = scene_bmp->ct - edge->top;
	    edge->top = scene_bmp->ct;
	    edge->x += edge->dx * gap;
	    _clip_polygon_segment(&edge->dat, gap, poly->flags);
	 }

	 if (edge->bottom >= edge->top) {
	    scene_inact = add_edge(scene_inact, edge, FALSE);
	    edge++;
            scene_nedge++;
	 }
      }
   }
   return 0;
}



/* scene_segment:
 * Draws a piece of the scanline, corresponding to a polygon. The start and
 * end values for x are taken from e01 and e02, the polygon style from poly.
 */
static void scene_segment(POLYGON_EDGE *e01, POLYGON_EDGE *e02,
                          POLYGON_INFO *poly)
{
   int x, w, gap, flags;
   int x01 = e01->x >> POLYGON_FIX_SHIFT;
   int x02 = e02->x >> POLYGON_FIX_SHIFT;
   float z01 = e01->dat.z;
   POLYGON_EDGE *e1 = poly->left_edge;
   POLYGON_EDGE *e2 = poly->right_edge;
   POLYGON_SEGMENT *info = &poly->info, *dat1, *dat2;
   SCANLINE_FILLER drawer;

   if ((x01 < last_x) && (z01 < last_z)) x01 = last_x;
   if ((x01 >= x02) || (x02 <= scene_bmp->cl) || (x01 >= scene_bmp->cr))
      return;
   if (x01 < scene_bmp->cl) x01 = scene_bmp->cl;
   if (x02 > scene_bmp->cr) x02 = scene_bmp->cr;
   if (x01 >= x02) return;

   if (!e2) {
      e2 = e02;
      while (e2 && (e2->poly != poly)) {
        e2 = e2->next;
      }
      if (!e2) return;
      poly->right_edge = e2;
   }

   x = e1->x >> POLYGON_FIX_SHIFT;
   w = (e2->x >> POLYGON_FIX_SHIFT) - x;
   if (w <= 0) return;

   flags = poly->flags;
   dat1 = &e1->dat;
   dat2 = &e2->dat;

   if (flags & INTERP_FLAT) {
      info->c = poly->color;
   } else if (flags & INTERP_1COL) {
      info->c = dat1->c;
      info->dc = (dat2->c - dat1->c) / w;
   }

   if (flags & INTERP_3COL) {
      info->r = dat1->r;
      info->g = dat1->g;
      info->b = dat1->b;
      info->dr = (dat2->r - dat1->r) / w;
      info->dg = (dat2->g - dat1->g) / w;
      info->db = (dat2->b - dat1->b) / w;
   }

   if (flags & INTERP_FIX_UV) {
      info->u = dat1->u + (1<<15);
      info->v = dat1->v + (1<<15);
      info->du = (dat2->u - dat1->u) / w;
      info->dv = (dat2->v - dat1->v) / w;
   }

   info->z = dat1->z;
   info->dz = (dat2->z - dat1->z) / w;

   if (flags & INTERP_FLOAT_UV) {
      info->fu = dat1->fu;
      info->fv = dat1->fv;
      info->dfu = (dat2->fu - dat1->fu) / w;
      info->dfv = (dat2->fv - dat1->fv) / w;
   }

   if (x < x01) {
      gap = x01 - x;
      x = x01;
      w -= gap;
      _clip_polygon_segment(info, gap, poly->flags);
   }

   if (x + w > x02)
      w = x02 - x;
      if (w <= 0) return;

   if ((flags & OPT_FLOAT_UV_TO_FIX) && (info->dz == 0)) {
      info->u = info->fu/info->z;
      info->v = info->fv/info->z;
      info->du = info->dfu/info->z;
      info->dv = info->dfv/info->z;
      drawer = poly->alt_drawer;
   } else {
      drawer = poly->drawer;
   }

   color_map = poly->cmap;
   _blender_alpha = poly->alpha;
   if (flags & INTERP_BLEND) {
      _blender_col_15 = poly->b15;
      _blender_col_16 = poly->b16;
      _blender_col_24 = poly->b24;
      _blender_col_32 = poly->b32;
   }
   if (drawer == _poly_scanline_dummy) {
      if (flags & INTERP_THRU) drawing_mode(DRAW_MODE_TRANS, NULL, 0, 0);
      hline(scene_bmp, x, scene_y, x+w, poly->color);
      if (flags & INTERP_THRU) solid_mode();
   } else {
      #ifdef USE_ZBUF
      if (flags & INTERP_ZBUF) {
         _zbuf_addr = bmp_write_line(zbuf_bmp, scene_y) + x * 4;
      }
      #endif
      drawer(scene_addr + x * ((bitmap_color_depth(scene_bmp) + 1) >> 3),
             w, info);
   }
}



/* scene_trans_seg:
 * Checks whether p0 is visible either directly or through transparent
 * polygons on top of it. If it's visible, draws all the visible stack
 * with x values from e1 and e2. At entry, p is the top polygon.
 * Returns nonzero if something was drawn.
 */
static int scene_trans_seg(POLYGON_EDGE *e1, POLYGON_EDGE *e2,
                           POLYGON_INFO *p0, POLYGON_INFO *p)
{
   int c;

   if (!p) return 0;
   c = 1;
   while (1) {
      if (p == p0) c = 0;
      if (!(p->flags & INTERP_THRU) || !p->next) break;
      p = p->next;
   }
   if (c) return 0;

   /* p is first opaque or the very last */
   while (p) {
      scene_segment(e1, e2, p);
      p = p->prev;
   }
   return 1;
}



/* scene_render:
 * Renders all the specified scene_polygon3d()'s on the bitmap passed to
 * scene_start(). Rendering is done one scanline at a time, with no pixel
 * being processed more than once. Note that between scene_start() and
 * scene_render() you shouldn't change the clip rectangle of the destination
 * bitmap. Also, all the textures passed to scene_polygon3d() are stored
 * as pointers only and actually used in scene_render().
 */
void scene_render()
{
   int p, old87;
   POLYGON_EDGE *edge, *start_edge = NULL;
   POLYGON_EDGE *active_edges = NULL, *last_edge = NULL;
   POLYGON_INFO *active_poly = NULL;

   scene_cmap = color_map;
   scene_alpha = _blender_alpha;
   solid_mode();
   /* set fpu to single-precision, truncate mode */
   old87 = _control87(PC_24 | RC_CHOP, MCW_PC | MCW_RC);
   _farsetsel(scene_bmp->seg);

   for (p=0; p<scene_npoly; p++) {
      scene_poly[p].inside = 0;
   }

   /* for each scanline in the clip region... */
   for (scene_y=scene_bmp->ct; scene_y<scene_bmp->cb; scene_y++) {
      scene_addr = bmp_write_line(scene_bmp, scene_y);

      /* check for newly active edges */
      edge = scene_inact;
      while ((edge) && (edge->top == scene_y)) {
	 POLYGON_EDGE *next_edge = edge->next;
	 scene_inact = remove_edge(scene_inact, edge);
	 active_edges = add_edge(active_edges, edge, TRUE);
	 edge = next_edge;
      }

      /* no edges on this line */
      if (!active_edges) continue;

      /* fill the scanline */
      last_x = -10000;
      last_z = 0.0;
      for (edge=active_edges; edge; edge=edge->next) {
         int x = edge->x >> POLYGON_FIX_SHIFT;
         POLYGON_INFO *poly = edge->poly;

         /* one polygon changes status */
         poly->inside = 1 - poly->inside;
         if (poly->inside) {
            POLYGON_INFO *pos = active_poly;
            POLYGON_INFO *prev = NULL;

            poly->left_edge = edge;
            poly->right_edge = NULL;

            /* find it's place in the list */
            while (pos && far_z(x, scene_y, edge, pos)) {
               prev = pos;
               pos = pos->next;
            }
            /* poly overlaps pos. Was pos visible ? */
            if (scene_trans_seg(start_edge, edge, pos, active_poly)) {
               start_edge = edge;
            }
            /* link */
            poly->next = pos;
            poly->prev = prev;
            if (pos) pos->prev = poly;
            if (prev) prev->next = poly;
            else {
               start_edge = edge;
               active_poly = poly;
            }
         } else {
            poly->right_edge = edge;
            /* poly ends here. Was it visible ? */
            if (scene_trans_seg(start_edge, edge, poly, active_poly)) {
               start_edge = edge;
               if (x > last_x) {
                  last_x = x;
                  last_z = edge->dat.z;
               }
            }
            /* unlink */
            if (poly->next) poly->next->prev = poly->prev;
            if (poly->prev) poly->prev->next = poly->next;
            else {
               active_poly = poly->next;
            }
         }
         /* prepare for backward scan */
         last_edge = edge;
      }

      /* update edges, remove dead ones, re-sort */
      edge = last_edge;
      active_edges = NULL;
      while (edge) {
	 POLYGON_EDGE *prev_edge = edge->prev;
	 if (scene_y < edge->bottom) {
            POLYGON_SEGMENT *dat = &edge->dat;
            int flags = edge->poly->flags;

	    edge->x += edge->dx;
            dat->z += dat->dz;

	    if (flags & INTERP_1COL)
	       dat->c += dat->dc;

            if (flags & INTERP_3COL) {
	       dat->r += dat->dr;
	       dat->g += dat->dg;
	       dat->b += dat->db;
	    }

	    if (flags & INTERP_FIX_UV) {
	       dat->u += dat->du;
	       dat->v += dat->dv;
	    }

	    if (flags & INTERP_FLOAT_UV) {
	       dat->fu += dat->dfu;
	       dat->fv += dat->dfv;
	    }

            active_edges = add_edge(active_edges, edge, TRUE);
	 }
	 edge = prev_edge;
      }
   }

   /* reset fpu mode */
   _control87(old87, MCW_PC | MCW_RC);
   color_map = scene_cmap;
   _blender_alpha = scene_alpha;
   solid_mode();

   /* mark the tables as full */
   scene_nedge = scene_maxedge;
   scene_npoly = scene_maxpoly;
}

